#include "feature.h"

#include <list>
#include "hdrimage.h"
#include <boost/shared_ptr.hpp>
#include <boost/math/special_functions/erf.hpp>
#include "func.h"
#include "dof.h"
#include "feature.h"
#include "featurefuziness.h"
#include "featurenoise.h"
#include "hdrimagefunctor.h"
#include "featureirrtosignal.h"

#ifdef _WIN32
#include <windows.h>
#else
#error alternative to GetTickCount???
#endif










void FeatureNoise::featureFtoF (PtrHDRImage3f in, PtrHDRImage3f& out)
{
	out = in;
	HDRIFunctorNoise hdriNoise (m_func, m_fmin, m_fmax);
	out->applyFunctor (hdriNoise);
}


void FeatureIrrToSignal::featureFtoF (PtrHDRImage3f in, PtrHDRImage3f& out)
{
	out = in;
	HDRIFunctorApplyFunc hdriFunctor (m_func);
	out->applyFunctor (hdriFunctor);
}




FeatureFuziness::FeatureFuziness (PtrFuncFtoF func)
{
	m_func = func;
	m_output = PtrHDRImage3f(new HDRImage3f);
	m_depth = PtrHDRImage1f(new HDRImage1f);
		
	m_func = PtrFuncFtoF (new FuncFuzinessConverter (func));
	m_dof = createLinGatherDOF (m_func);
}

void FeatureFuziness::featureFtoF (PtrHDRImage3f in, PtrHDRImage3f &out)
{
	initDepth (m_depth.get(), in->getSize());
	m_output->setSize (in->getSize(), in->getAppendSize());

	FuncFuzinessConverter* f = dynamic_cast<FuncFuzinessConverter*> (m_func.get());
	assert (f);
	f->setSize (in->getSize());

	m_dof->doDOF (in, m_depth, m_output);
	out = m_output;
}