#include "xml.h"
#include "hdrimage.h"
#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMDocumentType.hpp>
#include <xercesc/dom/DOMElement.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMNodeIterator.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMText.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <xercesc/util/XMLString.hpp>


using namespace xercesc;


#define DEF(SZ) XMLCh* sz##SZ = XMLString::transcode (#SZ)
#define REL(SZ) XMLString::release (&(sz##SZ))


bool XMLbase::checkAttribute (xercesc::DOMNode* node, string name, string value)
{
	XMLCh* xmlname = XMLString::transcode (name.c_str());
	XMLCh* xmlvalue = XMLString::transcode (value.c_str());
	for (uint i = 0; i < node->getAttributes()->getLength(); i++)
	{
		for (uint i = 0; i < node->getAttributes()->getLength(); i++)
		{
			const XMLCh* attr = node->getAttributes()->item(i)->getNodeName();
			const XMLCh* content = node->getAttributes()->item(i)->getTextContent();
			
			if (0 == XMLString::compareIString (xmlname, attr)
				&& 0 == XMLString::compareIString (xmlvalue, content))
			{
				XMLString::release (&xmlvalue);
				XMLString::release (&xmlname);
				return true;
			}
		}
	}
	XMLString::release (&xmlvalue);
	XMLString::release (&xmlname);
	return false;
}

xercesc::DOMNode* XMLbase::getChild (std::string name, xercesc::DOMNode* parent)
{
	DOMNode* ret = NULL;
	XMLCh* xmlname = XMLString::transcode (name.c_str());
	
	xercesc::DOMElement* element = static_cast<DOMElement*> (parent);
	if (! element)
		throwInvalid ("This node is not element");

	DOMNodeList* list = element->getElementsByTagName (xmlname);
	XMLString::release (&xmlname);

	if (! list->getLength ())
		throwNotFound (name);

	return list->item(0);
}
xercesc::DOMNode* XMLbase::getChildNotThrowNotFound (std::string name, xercesc::DOMNode* parent)
{
	DOMNode* ret = NULL;
	XMLCh* xmlname = XMLString::transcode (name.c_str());
	
	xercesc::DOMElement* element = static_cast<DOMElement*> (parent);
	if (! element)
		throwInvalid ("This node is not element");

	DOMNodeList* list = element->getElementsByTagName (xmlname);
	XMLString::release (&xmlname);

	if (! list->getLength ())
		NULL;

	return list->item(0);
}
bool XMLbase::getAttributeBool (xercesc::DOMNode* node, string name)
{
	bool btrue = checkAttribute (node, name, "true");
	bool bfalse = checkAttribute (node, name, "false");
	if (btrue != bfalse)
		return btrue;
	
	string msg = "attribute 'bool ";
	msg += name;
	msg += "'";
	throwInvalid (msg);
	return false;
}
string XMLbase::getAttributeString (xercesc::DOMNode* node, string name)
{
	XMLCh* xmlname = XMLString::transcode (name.c_str());
	for (uint i = 0; i < node->getAttributes()->getLength(); i++)
	{
		//for (uint i = 0; i < node->getAttributes()->getLength(); i++)
		//{
		const DOMNode* attrnode = node->getAttributes()->item(i);
		const DOMAttr* attr = static_cast<const DOMAttr*> (attrnode);
		
		const XMLCh* attrname = attr->getName ();
			
		if (0 == XMLString::compareIString (xmlname, attrname))
		{
			XMLString::release (&xmlname);
			const XMLCh* xmlvalue = node->getAttributes()->item(i)->getTextContent();
			char* value = XMLString::transcode (xmlvalue);
			string strval = value;
			delete [] (value);
			return strval;
		}
		//}

	}
	XMLString::release (&xmlname);
	throwNotFound (name);
	return "";
}

int XMLbase::getAttributeInt (DOMNode* node, string name)
{
	string str = getAttributeString (node, name);
	int i;
	std::stringstream ss(str); 
  
	if( (ss >> i).fail() )
	{
		string msg = "attribute 'int ";
		msg += name;
		msg += "'";
		throwInvalid (msg);
	}     
	return i;
}

uint XMLbase::getAttributeUInt (xercesc::DOMNode* node, string name)
{
	string str = getAttributeString (node, name);
	uint i;
	std::stringstream ss(str); 
  
	if( (ss >> i).fail() )
	{
		string msg = "attribute 'uint ";
		msg += name;
		msg += "'";
		throwInvalid (msg);
	}     
	return i;
}

float XMLbase::getAttributeFloat (xercesc::DOMNode* node, string name)
{

	string str = getAttributeString (node, name);
	float i;
	std::stringstream ss(str); 
  
	if( (ss >> i).fail() )
	{
		string msg = "attribute 'float ";
		msg += name;
		msg += "'";
		throwInvalid (msg);
	}     
	return i;
}

XMLfuncInter::XMLfuncInter(string camera, string version, xercesc::DOMNode* node):
		XMLfunc (camera, version)
{
	xercesc::DOMElement* elem = dynamic_cast <xercesc::DOMElement*> (node);
	DEF(point2);
	xercesc::DOMNodeList* list = elem->getElementsByTagName (szpoint2);
	m_points = new float2 [list->getLength()+2];
	for (uint i = 0; i < list->getLength(); i++)
	{
		xercesc::DOMNode* node = list->item(i);
		float2 v = getFloat2 (node);
		m_points[i+1] = v;
	}
	m_points[0] = float2 (-FLT_MAX, 0);
	m_points[list->getLength()+1] = float2 (FLT_MAX, 0);
	REL(point2);
}

template <uint Degree>
XMLfuncPolynome<Degree>::XMLfuncPolynome(string camera, string version, xercesc::DOMNode* node):
		XMLfunc (camera, version)
{
	xercesc::DOMElement* elem = dynamic_cast <xercesc::DOMElement*> (node);
	DEF(value);
	xercesc::DOMNodeList* list = elem->getElementsByTagName (szvalue);
	if (list->getLength() != Degree+1)
		throwInvalid ("'degree' != count of the 'value's");
	for (uint i = 0; i < list->getLength(); i++)
	{
		xercesc::DOMNode* node = list->item(i);
		const XMLCh* text =  node->getTextContent ();
		char* sz = XMLString::transcode (text);
		std::stringstream ss (sz);
		float f;
		if( (ss >> f).fail() )
		{
			throwInvalid ("a 'value' in the func");
		} 
		delete [] (sz);
		int index = Degree - i;
		m_coef[index] = f;
	}
	REL(value);
}

void XMLbase::getGrid (xercesc::DOMNode* node, HDRImage1f& gridX, HDRImage1f& gridY)
{
	uint2 size; //= getPoint2D (node);
	size.x = getAttributeInt (node, "countX");
	size.y = getAttributeInt (node, "countY");
	xercesc::DOMElement* elem = dynamic_cast <xercesc::DOMElement*> (node);
	DEF(point2);
	xercesc::DOMNodeList* list = elem->getElementsByTagName (szpoint2);
	if (list->getLength() != size.getArea())
		throwInvalid ("'size of grid' != count of the 'point2's");
	gridX.setSize (size);
	gridY.setSize (size);
	for (uint i = 0; i < list->getLength(); i++)
	{
		DOMNode* node = list->item (i);
		float2 f2 = getFloat2 (node);
		gridX.setPixel (uint2 (i%size.x, i/size.x), f2.x);
		gridY.setPixel (uint2 (i%size.x, i/size.x), f2.y);
	}
	REL(point2);
}


float XMLbase::getFloat (xercesc::DOMNode* node)
{
	float x;
	x = getAttributeFloat (node, "x");
	return x;
}

float2 XMLbase::getFloat2  (xercesc::DOMNode* node)
{
	float x, y;
	x = getAttributeFloat (node, "x");
	y = getAttributeFloat (node, "y");
	return float2 (x, y); 
}

float3 XMLbase::getFloat3  (xercesc::DOMNode* node)
{
	float x, y, z;
	x = getAttributeFloat (node, "x");
	y = getAttributeFloat (node, "y");
	z = getAttributeFloat (node, "z");
	return float3 (x, y, z); 
}

PtrXMLfunc XMLbase::getChildFunc (xercesc::DOMNode* node)
{
	DOMNode* nodefunc = getChild ("func", node);
	if ((getAttributeString (nodefunc, "type")) == "lininter")
		return PtrXMLfunc (new XMLfuncInter (m_strCamera, m_strVersion, nodefunc));
	else if (getAttributeString (nodefunc, "type") == "polynomial")
	{
		switch (getAttributeInt (nodefunc, "degree"))
		{
#define CASE_N(n) case n: return PtrXMLfunc (new XMLfuncPolynome<n> (m_strCamera, m_strVersion, nodefunc)); break;
		CASE_N(0)
		CASE_N(1)
		CASE_N(2)
		CASE_N(3)
		CASE_N(4)
		CASE_N(5)
		CASE_N(6)
		CASE_N(7)
		CASE_N(8)
		CASE_N(9)
		CASE_N(10)
		CASE_N(11)
		CASE_N(12)
		CASE_N(13)
		CASE_N(14)
		CASE_N(15)
		CASE_N(16)
#undef CASE_N
		default:
			throwInvalid ("degree of the function is not implemented");
		}
	}
	else 
		throwInvalid ("type of the function");
	PtrXMLfunc null;
	return null;
}


XMLchromAbberation::XMLchromAbberation (string camera, string version, DOMNode* node):
	XMLbase (camera, version, "chromatic_abberation")
{
	node = getChild ("lens", node);
	node = getChild ("chromatic_abberation", node);
	m_enabled = getAttributeBool (node, "enabled");

	m_sizeOfGrid.x = getAttributeUInt (getChild ("sizeOfGrid", node), "x");
	m_sizeOfGrid.y = getAttributeUInt (getChild ("sizeOfGrid", node), "y");

	DOMNode* nodered = getChild ("red_cyan", node);
	m_redcyan = getChildFunc (nodered);

	DOMNode* nodeblue = getChild ("blue_yellow", node);
	m_blueyellow = getChildFunc (nodeblue);
}


void XMLgeometricDistortion::processRadialDistortion (xercesc::DOMNode* node)
{
	uint2 gridSize;
	{
		gridSize.x = getAttributeUInt (getChild ("sizeOfGrid", node), "x"); 
		gridSize.y = getAttributeUInt (getChild ("sizeOfGrid", node), "y"); 
	}
	PtrXMLfunc func = getChildFunc (node);
	
	float3 mult[2];
	{
		xercesc::DOMNode* param;
		param = getChild ("param", node);
		mult[0].x = getAttributeFloat (param, "alpha");
		mult[0].y = getAttributeFloat (param, "gamma");
		mult[0].z = getAttributeFloat (param, "u0");
		mult[1].x = 0;
		mult[1].y = getAttributeFloat (param, "beta");
		mult[1].z = getAttributeFloat (param, "v0");
	}


	m_type = TYPE_NET;
	//
	m_gridX = PtrHDRImage1f (new HDRImage1f);
	m_gridY = PtrHDRImage1f (new HDRImage1f);

	m_gridX->setSize (gridSize);
	m_gridY->setSize (gridSize);
	for (uint y = 0; y < gridSize.y; y++)
		for (uint x = 0; x < gridSize.x; x++)
		{
			float rx = x/((float)gridSize.x-1)*2-1;
			float ry = y/((float)gridSize.y-1)*2-1;

			float prevr;
			float r = prevr = sqrtf(rx*rx+ry*ry) / sqrtf(2);
			r = (*func) (r);
			
			rx *= r/prevr;
			ry *= r/prevr;

			// rx = <-1;1> 			
			//use coef:
			
			float _rx = rx*mult[0].x + ry*mult[0].y + mult[0].z;
			float _ry = rx*mult[1].x + ry*mult[1].y + mult[1].z;

			rx = (_rx+1)/2;
			ry = (_ry+1)/2;

			//set to grid:
			m_gridX->setPixel(uint2(x,y), rx);
			m_gridY->setPixel(uint2(x,y), ry);
		}
}


XMLgeometricDistortion::XMLgeometricDistortion (string camera, string version, xercesc::DOMNode* node):
	XMLbase (camera, version, "geometric_distortion")
{
	node = getChild ("lens", node); 
	node = getChild ("geometric_distortion", node);
	m_enabled = getAttributeBool (node, "enabled");
	string strtype = getAttributeString (node, "type");
	if (strtype == "net")
	{
		// type == net
		m_type = TYPE_NET;
		//
		m_gridX = PtrHDRImage1f (new HDRImage1f);
		m_gridY = PtrHDRImage1f (new HDRImage1f);

		HDRImage1f& gridX = *(m_gridX.get());
		HDRImage1f& gridY = *(m_gridY.get());
		XMLbase::getGrid (getChild ("grid", node), gridX, gridY);
	}
	else if (strtype == "radial")
	{
		processRadialDistortion (node);
	}
	else
		throwInvalid ("type");
	
}
			
XMLvignetting::XMLvignetting (string camera, string version, xercesc::DOMNode* node):
	XMLbase (camera, version, "vignetting")
{
	node = getChild ("lens", node);
	node = getChild ("vignetting", node);
	m_enabled = getAttributeBool (node, "enabled");
	string strtype = getAttributeString (node, "type");
	if (0 == strtype.compare ("regular"))
		m_type = TYPE_REGULAR;
	else if (0 == strtype.compare ("irregular"))
		m_type = TYPE_IRREGULAR;
	else 
		throwInvalid ("type");
	if (m_type == TYPE_IRREGULAR)
	{
		DOMNode* nodemask = getChild ("mask", node); 
		m_irregFilepath = getAttributeString (nodemask, "filepath");
	}
	if (m_type == TYPE_REGULAR)
	{
		m_regFunc = getChildFunc (node);
	}
}

XMLfuziness::XMLfuziness (string camera, string version, xercesc::DOMNode* node):
	XMLbase (camera, version, "lens_sharpness")
{
	node = getChild ("lens", node);
	node = getChild ("lens_sharpness", node);
	m_enabled = getAttributeBool (node, "enabled");
	m_func = getChildFunc (node);
}

XMLlensflare::XMLlensflare (string camera, string version, xercesc::DOMNode* node):
	XMLbase (camera, version, "lens_flare")
{
	node = getChild ("lens", node);
	node = getChild ("lens_flare", node);
	m_enabled = getAttributeBool (node, "enabled");
	string strtype = getAttributeString (node, "type");
	if (strtype != "set")
		throwInvalid ("type");

	string setname = getAttributeString (getChild("set", node), "name");
	if (setname != "simpleinscreen")
		throwInvalid ("set/name");

	m_envmapangle.x = getAttributeFloat (getChild("envmap", node), "angleX");
	m_envmapangle.y = getAttributeFloat (getChild("envmap", node), "angleY");
	m_sizeofenvmap.x = getAttributeUInt (getChild("envmap", node), "sizeX");
	m_sizeofenvmap.y = getAttributeUInt (getChild("envmap", node), "sizeY");
	
	m_fovangle.x = getAttributeFloat (getChild("fov", node), "angleX");
	m_fovangle.y = getAttributeFloat (getChild("fov", node), "angleY");
	
	m_starMult  = getAttributeFloat (getChild("params", node), "starMult");
	m_csMult1 = getAttributeFloat (getChild("params", node), "csMult1");
	m_csMult2 = getAttributeFloat (getChild("params", node), "csMult2");
}

XMLmotionblur::XMLmotionblur (string camera, string version, DOMNode* node):
	XMLbase (camera, version, "motion_blur")
{
	node = getChild ("motion_blur", node);
	m_enabledParameters = getAttributeBool (node, "enabled") &&
		(getAttributeString(node, "type") == "parameters");

	numIterations = getAttributeUInt (getChild("iterations", node), "number");
	node = getChild ("motion_blur_param", node);

	deltaCameraPos = getFloat3(getChild("point3", getChild ("deltapos", node)));
	deltaTarget = getFloat2 (getChild("point2", getChild ("deltatarget", node)));
	targetDist = getFloat (getChild("float", getChild ("distancetarget", node)));
	fovAngle = getFloat2 (getChild("point2", getChild ("fovangle", node)));
}

XMLdepthoffield::XMLdepthoffield (string camera, string version, DOMNode* node):
	XMLbase (camera, version, "depth_of_field")
{
	node = getChild ("depth_of_field", node);
	m_enabled = getAttributeBool (node, "enabled");

	string type = getAttributeString (getChild ("type", node), "name");
	if (type == "gather")
		m_type = TYPE_GATHER;
	else if (type == "diffusion")
	{
		m_type = TYPE_DIFFUSION;
		m_numIterations = getAttributeInt (getChild ("type", node), "iterations");
	}
	else
		throwInvalid ("type name");

	type = getAttributeString (getChild ("aperture", node), "type");
	if (type == "fixed")
	{
		m_apertureIsFixed = true;
		m_apertureValue = getAttributeFloat (getChild ("aperture", node), "value");
	}
	else if (type == "adapt")
	{
		m_apertureIsFixed = false;
	}
	else
		throwInvalid ("aperture type");
	
	type = getAttributeString (getChild ("distance", node), "type");
	if (type == "fixed")
	{
		//m_focusIsFixed = true;
		m_focusDist = getAttributeFloat (getChild ("distance", node), "depth");
	}
	else
		throwInvalid ("distance");

	DOMNode* params = getChild ("parameters", node);
	m_sensorSize = getAttributeFloat (params, "sensorSize");
	m_focalLen = getAttributeFloat (params, "focalLen");
}

XMLaperture::XMLaperture (string camera, string version, xercesc::DOMNode* node):
	XMLbase (camera, version, "aperture")
{
	node = getChild ("lens", node);
	node = getChild ("aperture", node);

	string type = getAttributeString (node, "type");
	if (type == "fixed")
		m_type = TYPE_FIXED;
	else if (type == "adapt")
		m_type = TYPE_ADAPT;
	else
		throwInvalid ("type");

	m_fnum = getAttributeFloat (node, "fnum");
	m_brightness = getAttributeBool (node, "brightness");

	if (m_type == TYPE_ADAPT)
	{
		node = getChild ("adapt", node);
		m_param.middleGray = getAttributeFloat (node, "middlegray");
		m_param.minFnum = getAttributeFloat (node, "minFnum");
		m_param.maxFnum = getAttributeFloat (node, "maxFnum");
		m_param.maxSpeed = getAttributeFloat (node, "maxSpeed");
		m_param.Kp = getAttributeFloat (node, "PIregKp");
		m_param.Ki = getAttributeFloat (node, "PIregKi"); 
	}
}


XMLcolorfilterarray::XMLcolorfilterarray (string camera, string version, xercesc::DOMNode* node):
	XMLbase (camera, version, "color_filter_array")
{
	DOMNode* basenode = getChild ("color_filter_array", node);
	m_enabled = getAttributeBool (basenode, "enabled");
	for (uint i = 0; i < 3; i++)
	{
		string color;
		FeatureColorFilterArrayParam* p; 
		switch (i)
		{
		case 0: color="red"; p = &m_red;break;
		case 1: color="green"; p = &m_green;break;
		case 2: color="blue"; p = &m_blue;break;
		}
		node = getChild(color, basenode);
		{//mask:
			DOMNode* nodemask = getChild("mask", node);
			uint x = getAttributeUInt (nodemask, "x");
			uint y = getAttributeUInt (nodemask, "y");
			string mask = getAttributeString (nodemask, "mask");
			p->mask->setSize (uint2 (x,y));
			if (mask.length() != x*y)
				throwInvalid ("mask length");
			for (uint i = 0; i < mask.length(); i++)
			{
				float value;
				char c = mask[i];
				if (c == '0')
					value = 0;
				else if (c == '1')
					value = 1;
				else
					throwInvalid ("mask");

				p->mask->setPixel (uint2(i%x, i/x), value);	
			}
		}
		{ // kernel:
			DOMNode* nodekernel = getChild ("kernel", node);
			int ix = getAttributeInt (nodekernel, "offsetX");
			int iy = getAttributeInt (nodekernel, "offsetY");
			p->kernelOff = int2 (ix, iy);

			uint x = getAttributeUInt (nodekernel, "sizeX");
			uint y = getAttributeUInt (nodekernel, "sizeY");
			p->kernel->setSize (uint2 (x, y));

			xercesc::DOMElement* elem = dynamic_cast <xercesc::DOMElement*> (node);
			DEF(value);
			xercesc::DOMNodeList* list = elem->getElementsByTagName (szvalue);
			for (uint i = 0; i < list->getLength(); i++)
			{
				xercesc::DOMNode* node = list->item(i);
				float f = getFloat (node);
				p->kernel->setPixel(uint2(i%x, i/x), f);
			}
			REL(value);
		}
	}
}


XMLcamera::XMLcamera (string camera, string version, DOMNode* node):
	XMLbase (camera, version), 
	m_vignetting (new XMLvignetting (camera, version, node)),
	m_distortion (new XMLgeometricDistortion (camera, version, node)),
	m_chromabberation (new XMLchromAbberation (camera, version, node)),
	m_fuziness(new XMLfuziness (camera, version, node)),
	m_lensflare(new XMLlensflare (camera, version, node)),
	m_motionblur(new XMLmotionblur (camera, version, node)),
	m_depthoffield(new XMLdepthoffield (camera, version, node)),
	m_aperture(new XMLaperture (camera, version, node)),
	m_cfa(new XMLcolorfilterarray (camera, version, node))
{
	DOMNode* basenode = node;
	node = getChild ("sensor", node);
	m_funcIrrToSignal = getChildFunc (getChild("irradiance_to_signal_conversion", node));
	m_funcSignalToNoise = getChildFunc (getChild("signal_to_noise_ratio", node));
	m_isNoiseEnabled = getAttributeBool (getChild("signal_to_noise_ratio", node), "enabled");
	m_noiseMin = getAttributeFloat (getChild("limit", node), "min");
	m_noiseMax = getAttributeFloat (getChild("limit", node), "max"); 

	node = getChild ("order", basenode);
	xercesc::DOMElement* elem = dynamic_cast <xercesc::DOMElement*> (node);
	DEF(feature);
	xercesc::DOMNodeList* list = elem->getElementsByTagName (szfeature);
	for (uint i = 0; i < list->getLength(); i++)
	{
		string name = getAttributeString (list->item(i), "name");
		
		efeature f = NONE;
#define DEF_F(cstr,cval) {if (name == cstr) f = cval;}
		DEF_F ("dof", DOF);
		DEF_F ("blur", BLUR);
		DEF_F ("motion", MOTION);
		DEF_F ("lensflare", LENSFLARE);
		DEF_F ("distortion", DIST);
		DEF_F ("chromabber", CHROMABBER);
		DEF_F ("vignetting", VIGNETTING);
		DEF_F ("adaptdo", ADAPTDO);
		DEF_F ("adaptget", ADAPTGET);
		DEF_F ("clamp", CLAMP);
		DEF_F ("irrtosignal", IRRTOSIGNAL);
		DEF_F ("noise", NOISE);
		DEF_F ("cfa", CFA);
#undef DEF_F
		if (f == NONE)
			throwInvalid ("order/feature");
		m_vecFeatures.push_back (f);
	}
	REL(feature);

}

XMLparser::XMLparser (string filepath, string camera, string version, string input, string output):
	XMLbase (camera, version)
{
	try
	{
		XMLPlatformUtils::Initialize();  // Initialize Xerces infrastructure
	}
	catch (XMLException& e)
	{
		char* message = XMLString::transcode (e.getMessage());
		string str = message; 
		XMLString::release (&message);
		throwException (str);
	}
	XercesDOMParser parser;
	parser.setValidationScheme (XercesDOMParser::Val_Never);
	parser.setDoNamespaces (false);
	parser.setDoSchema (false);
	parser.setLoadExternalDTD (false);

	parser.parse (filepath.c_str());
	DOMDocument* doc = parser.getDocument();
	if (! doc)
		throwException ("xml file cannot be loaded");
	
	{
		DEF(camera);
		DEF(name);
		DEF(version);

		DOMElement* root = doc->getDocumentElement ();
		DOMNodeList* cameras = root->getElementsByTagName (szcamera);

		bool valid = false;
		for (uint i = 0; i < cameras->getLength(); i++)
		{
			DOMNode* node = cameras->item(i);
			if (checkAttribute (node, "name", camera) && checkAttribute (node, "version", version))
			{
				m_camera = boost::shared_ptr<XMLcamera> (new XMLcamera (camera, version, node));
				valid = true;
				break;
			}
		}
		REL(camera);
		REL(name);
		REL(version);

		if (! valid)
			throwNotFound ("camera");
	}

	{
		DEF(input);

		DOMElement* root = doc->getDocumentElement ();
		DOMNodeList* inputs = root->getElementsByTagName (szinput);

		bool valid = false;
		for (uint i = 0; i < inputs->getLength(); i++)
		{
			DOMNode* node = inputs->item(i);
			if (checkAttribute (node, "name", input))
			{
				m_input = PtrXMLinput (new XMLinput (node));
				valid = true;
				break;
			}
		}
		REL(input);

		if (! valid)
			throwNotFound ("input");
	}

	{
		DEF(output);

		DOMElement* root = doc->getDocumentElement ();
		DOMNodeList* inputs = root->getElementsByTagName (szoutput);

		bool valid = false;
		for (uint i = 0; i < inputs->getLength(); i++)
		{
			DOMNode* node = inputs->item(i);
			if (checkAttribute (node, "name", output))
			{
				m_output = PtrXMLoutput (new XMLoutput (node));
				valid = true;
				break;
			}
		}
		REL(output);

		if (! valid)
			throwNotFound ("input");
	}
}

XMLvideoExt::XMLvideoExt (xercesc::DOMNode* node):
	XMLbase ("", "", "input/video")
{
	start = getAttributeInt(node, "start");
	end = getAttributeInt(node, "end");
	delay = getAttributeInt(node, "delay");
	loop = getAttributeInt(node, "loop");
}

XMLinput::XMLinput (xercesc::DOMNode* node):
		XMLbase ("", "", "input")
{
	xercesc::DOMElement* elem = dynamic_cast <xercesc::DOMElement*> (node);
	DEF(file);
	xercesc::DOMNodeList* list = elem->getElementsByTagName (szfile);
	for (uint i = 0; i < list->getLength(); i++)
	{
		xercesc::DOMNode* node = list->item(i);
		files.push_back (XMLinputFile (node));
	}
	REL(file);

	colorMult = XMLbase::getAttributeFloat(getChild ("color", node), "mult");
	depthLinearize = XMLbase::getAttributeBool(getChild ("depth", node), "linearize");
	depthMult = XMLbase::getAttributeFloat(getChild ("depth", node), "mult");
	depthAdd = XMLbase::getAttributeFloat(getChild ("depth", node), "add");
	depthN = XMLbase::getAttributeFloat(getChild ("depth", node), "linearizeNear");
	depthF = XMLbase::getAttributeFloat(getChild ("depth", node), "linearizeFar");

	icosmult = XMLbase::getAttributeFloat(getChild ("depth", node), "icosmult");
	icosanglemult = XMLbase::getAttributeFloat(getChild ("depth", node), "anglemult");

	if (getChildNotThrowNotFound ("video", node))
		videoExt = PtrXMLvideoExt (new XMLvideoExt (getChild ("video", node)));
}


XMLoutputDemo::XMLoutputDemo (xercesc::DOMNode* node):
	XMLbase ("", "", "output")
{
	wndSize.x = getAttributeUInt (getChild ("window", node), "x");
	wndSize.y = getAttributeUInt (getChild ("window", node), "y");

	lightFilepath = "-=not used=-";
	lightEnabled = false;
	lightMult = 0;

	/*node = getChild ("light", node);
	lightFilepath = getAttributeString (node, "path");
	lightEnabled = getAttributeBool (node, "enabled");
	lightMult = getAttributeFloat (node, "mult");
*/
}
XMLoutput::XMLoutput (xercesc::DOMNode* node):	
	XMLbase ("", "", "output")
{
	string strtype = getAttributeString (node, "type");
	if (strtype == "demo")
	{
		type = TYPE_DEMO;
		m_demo = PtrXMLoutputDemo (new XMLoutputDemo (node));
	}
	else if (strtype == "file")
	{
		type = TYPE_FILE;
		m_file = PtrXMLoutputFile (new XMLoutputFile (node));
	}
}


XMLoutputFile::XMLoutputFile (DOMNode* base):
	XMLbase ("", "", "output")
{
	{
		DOMNode* node = getChild ("window", base);
		wndSize.x = getAttributeUInt (node, "x");
		wndSize.y = getAttributeUInt (node, "y");
	}

	{
		DOMNode* node = getChild ("file", base);
		fileType = getAttributeString (node, "type");
		if (fileType != "bmp")
			throwInvalid ("type");
		filePrefix = getAttributeString (node, "prefix");
		fileNumDigits = getAttributeUInt (node, "numDigits");
		fileSuffix = getAttributeString (node, "suffix");
	}

	{
		DOMNode* node = getChild ("iterations", base);
		itBegin = getAttributeUInt (node, "begin");
		itStep = getAttributeUInt (node, "step");
		itCount = getAttributeUInt (node, "count");
	}
}